/*
 *  Copyright (c) 2010-2013, Rutgers, The State University of New Jersey
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are met:
 *      * Redistributions of source code must retain the above copyright
 *        notice, this list of conditions and the following disclaimer.
 *      * Redistributions in binary form must reproduce the above copyright
 *        notice, this list of conditions and the following disclaimer in the
 *        documentation and/or other materials provided with the distribution.
 *      * Neither the name of the organization(s) stated above nor the
 *        names of its contributors may be used to endorse or promote products
 *        derived from this software without specific prior written permission.
 *  
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
 *  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 *  ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
 *  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * @file   mfsegmentor.cpp
 * @author wontoniii (bronzino@winlab.rutgers.edu), czhang
 * @date   July, 2011
 * @brief  Class that handles the transmission of data chunks, csyn and assoc/dassoc packets.
 *
 * Class that handles the transmission of data chunks, csyn and assoc/dassoc packets.
 */

#include <pcap.h>
#include <vector>
#include <errno.h>

#include <include/mfhopmsg.hh>
#include <include/mfgstarctrlmsg.hh>
#include <include/mfroutingheader.hh>
#include <include/mfmhomeextheader.hh>

#include "mfsegmentor.h"
#include "mftypes.h"
#include "mfdeviceinfo.h"
#include "mfchunkinfo.h"
#include "mfipproto.h"
#include "mfetherproto.h"
#include "mftimer.h"
#include "mflog.h"
#include "mfeventqueue.h"
#include "mfpacketsupport.h"

//TODO: redo this class (at least update parts that are new in stack

MF_Segmentor::MF_Segmentor() {
	mLastHopID = 0;
	mIsCsynSent = false;
	mCsynSeqNum = 0; //used by timer function
	sendingChunk = NULL;
}

MF_Segmentor::MF_Segmentor(MF_EventQueue *eq, pcap_t *handle, int devId, int t) : mTimer(eq, devId, t) {
	mLastHopID = 0;
	mIsCsynSent = false;
	mCsynSeqNum = 0; //used by timer function
	this->eq = eq;
	mPcapHandle = handle;
	sendingChunk = NULL;
}

MF_Segmentor::~MF_Segmentor() {

}

void MF_Segmentor::setPCAPHandle(pcap_t *handle){
	mPcapHandle = handle;
}

u_int MF_Segmentor::getNewHopID() {
	u_int newID = (mLastHopID + 1) % MAX_HOP_ID;
	mLastHopID = newID;
	return newID;
}

void MF_Segmentor::resetHopID(u_int id){
	mLastHopID = id;
}

u_int MF_Segmentor::updateCsynSeqNum() {
	uint newNum = (mCsynSeqNum + 1) % MAX_CSYN_SEQ_NUM;
	mCsynSeqNum = newNum;
	return newNum;
}

u_int MF_Segmentor::getCsynSeqNum() {
	return mCsynSeqNum;
}

bool MF_Segmentor::isSending(){
	return sendingChunk!=NULL;
}

void MF_Segmentor::procSend(MF_ChunkInfo *ci, MF_DeviceInfo *di) {
	MF_Log::mf_log(MF_DEBUG, "MF_Segmentor:procSend send new chunk");
	
	u_int hopID = getNewHopID();
	sendingChunk = ci;
	ci->putHopID(hopID);
	fillHeader(ci, di);
	MF_Log::mf_log(MF_TIME, "MF_Segmentor:procSend SEND_CSYN INIT chunk %lu interface %d", ci->getChunkID(), di->getID());
	sendCsyn(ci, di);
	mTimer.reset(ci->getChunkID());
}

bool MF_Segmentor::isAllSent(u_char *bitmap) {
	MF_Log::mf_log(MF_DEBUG, "MF_Segmentor:isAllSent check if chunk is all sent");
	u_int i = 0;
	bool ret = true;
	mNonAcked.clear();
	for (i = 0; i < sendingChunk->getChunkPktCnt(); i++) {
		if ((bitmap[i / 8] & (0x01 << (i % 8))) == 0) { //not received
			mNonAcked.push_back(i);
			ret = false;
		}
	}
	return ret;
}

void MF_Segmentor::sendData(MF_ChunkInfo *ci, MF_DeviceInfo *di) {
	MF_Log::mf_log(MF_DEBUG, "MF_Segmentor:sendData send data chunk");
	u_int i = 1;
	int ret = 0;
	struct hop_data *hop;
	vector <u_char *> *pList = ci->getPacketList();
	for (i = 0; i < mNonAcked.size(); i++) {
		hop = (struct hop_data *) MF_PacketSupport::getHopHeaderPtr((*pList)[i]);
		ret = pcap_inject(mPcapHandle, (void *) ((*pList)[mNonAcked[i]]), MF_PacketSupport::LOW_HEADER_LEN + ntohl(hop->pld_size));
		if (ret < 0) {
			MF_Log::mf_log(MF_ERROR, "MF_Segmentor::sendData(): pcap_inject() error, errno=%d", errno);
			return;
		}
#ifdef __OML__
		di->increaseOutBytesByInt(MF_PacketSupport::LOW_HEADER_LEN + ntohl(hop->pld_size));
#endif
	}
	MF_Log::mf_log(MF_DEBUG, "MF_Segmentor::sendData(): Sent %d packets. hop_ID = %d", mNonAcked.size(), ci->getHopID());
	MF_Log::mf_log(MF_TIME, "MF_Segmentor:procSend SEND_CSYN INIT chunk %lu interface %d", ci->getChunkID(), di->getID());
	sendCsyn(ci, di);
	mTimer.reset(ci->getChunkID());
	
#ifdef __OML__
	di->increaseOutPacketsByInt(mNonAcked.size());
	di->sendOMLUpdate();
#endif
}

void MF_Segmentor::fillHeader(MF_ChunkInfo *ci, MF_DeviceInfo *di) {
	u_int reminSize = ci->getChunkSize() + MF_PacketSupport::HIGH_HEADER_LEN;
	u_int seqNum = 0;
	u_int pldSize = 0;
	u_int i;
	vector <u_char *> *pList = ci->getPacketList();
	for (i = 0; i < ci->getChunkPktCnt(); i++) {
		pldSize = (reminSize > (u_int)MF_PacketSupport::MAX_PAYLOAD_SIZE) ?(u_int)MF_PacketSupport::MAX_PAYLOAD_SIZE : reminSize;
		MF_EtherProto::fillHeader((*pList)[i], 0, di->getSMAC().c_str(), di->getAPMAC().c_str());
		MF_IPProto::fillHeader((*pList)[i], ETH_HEADER_SIZE, pldSize + HOP_HEADER_SIZE, di->getSIP().c_str(), di->getAPIP().c_str());
		fillHopHeader((*pList)[i], ci->getHopID(), pldSize, seqNum);
		seqNum++;
		reminSize -= pldSize;
	}
}

void MF_Segmentor::fillHopHeader(u_char *buffer, u_int hopID, u_int pldSize, u_int seqNum) {
	struct hop_data *hop;
	hop = (struct hop_data *) MF_PacketSupport::getHopHeaderPtr(buffer);
	hop->type = htonl(DATA_PKT);
	hop->hop_ID = htonl(hopID);
	hop->pld_size = htonl(pldSize);
	hop->seq_num = htonl(seqNum);
}

void MF_Segmentor::sendCsyn(MF_ChunkInfo *ci, MF_DeviceInfo *di) {
	MF_Log::mf_log(MF_DEBUG, "MF_Segmentor:sendCsyn send csyn packet");
	u_char buffer[256];
	int ret = 0;
	struct csyn *csyn;
	csyn = (struct csyn *) MF_PacketSupport::getHopHeaderPtr(buffer);
	csyn->type = htonl(CSYN_PKT);
	csyn->hop_ID = htonl(ci->getHopID());
	csyn->chk_pkt_count = htonl(ci->getChunkPktCnt());
	MF_IPProto::fillHeader(buffer, ETH_HEADER_SIZE, sizeof(struct csyn), di->getSIP().c_str(), di->getAPIP().c_str());
	MF_EtherProto::fillHeader(buffer, 0, di->getSMAC().c_str(), di->getAPMAC().c_str());
	if ((ret = pcap_inject(mPcapHandle, (void *) buffer, ETH_HEADER_SIZE + IP_HEADER_SIZE + sizeof(struct csyn))) < 0) {
		MF_Log::mf_log(MF_ERROR, "pcap_sendpacket error, errno=%d", errno);
		return;
	}
	MF_Log::mf_log(MF_DEBUG, "MF_Segmentor: Sent CSYN hop_ID = %d, pkt_count = %d, if = %s", ci->getHopID(), ci->getChunkPktCnt(), di->getName().c_str());
	updateCsynSeqNum();
	mIsCsynSent = true;
	
#ifdef __OML__
	di->increaseOutPackets();
	di->increaseOutBytesByInt(ETH_HEADER_SIZE + IP_HEADER_SIZE + sizeof(struct csyn));
#endif
}

void MF_Segmentor::csynTO(u_int chunkID, MF_DeviceInfo *di){
	MF_Log::mf_log(MF_DEBUG, "MF_Segmentor:csynTO TO has triggered");
	if(sendingChunk!=NULL && sendingChunk->getChunkID() == mTimer.getChunkID() && mIsCsynSent){
		MF_Log::mf_log(MF_DEBUG, "MF_Segmentor:csynTO has to resend the csyn because it got lost");
		sendCsyn(sendingChunk, di);
		mTimer.reset(sendingChunk->getChunkID());
	}
	else{
		MF_Log::mf_log(MF_DEBUG, "MF_Segmentor:csynTO nothing to do. Should have not triggered");
	}
}

void MF_Segmentor::handleAck(u_int hop_ID, u_int pkt_count, u_char *bitmap, MF_DeviceInfo *di) {
	MF_Log::mf_log(MF_DEBUG, "MF_Segmentor:handleAck Recvd csyn-ack for hop_ID = %d, pkt_count = %d, if = %s", hop_ID, pkt_count, di->getName().c_str());
	//TODO remove the harcoded value and make it so that the bitmap is created only if debugging is active
	char tempBitmap[512];
	u_int n;
	for (n = 0; (n < (pkt_count - 1) / 8 + 1) && n < 512; n++) {
		sprintf(tempBitmap + n, "%x", bitmap[n]);
	}
	MF_Log::mf_log(MF_DEBUG, "MF_Segmentor:handleAck received bitmap: %s", tempBitmap);
	if ((mIsCsynSent == false) || (hop_ID != mLastHopID)) { //wrong state, return directly
		MF_Log::mf_log(MF_WARN, "MF_Segmentor:handleAck(): unexpected csyn-ack: expected hopID %d, got %d", mLastHopID, hop_ID);
		return;
	}
	
	mIsCsynSent = false;
	
	if(isAllSent(bitmap)) {
		mTimer.stop();
		MF_EvChkComplete *event = new MF_EvChkComplete(di->getID(), sendingChunk);
		sendingChunk = NULL;
		eq->add(event);
		return;
	}
	MF_Log::mf_log(MF_TIME, "MF_Segmentor:handleAck SEND_DATA INIT hop_id %lu", hop_ID);
	sendData(sendingChunk, di);
	MF_Log::mf_log(MF_TIME, "MF_Segmentor:handleAck SEND_DATA END hop_id %lu", hop_ID);
}
							
bool MF_Segmentor::sendAck(MF_ChunkInfo *ci, MF_DeviceInfo *di) {
	MF_Log::mf_log(MF_DEBUG, "MF_Segmentor:sendAck send ack packet for given chunk");
	u_char buffer[65536];// large enough for any size of bitmap
	memset((void *) buffer, 0xff, 65536);
	struct hop_data *hop;
	u_int pktCount = ci->getChunkPktCnt();
	struct csyn_ack *ack;
	ack = (struct csyn_ack *) MF_PacketSupport::getHopHeaderPtr(buffer);
	ack->type = htonl(CSYN_ACK_PKT);
	ack->hop_ID = htonl((u_int) (ci->getHopID()));
	ack->chk_pkt_count = htonl(pktCount);
	
	u_int i = 0;
	u_int j = 0;
	int ret;
	
	if (ci->isDelivered()) {// last ack get lost
		for (i = 0; i < (pktCount - 1) / 8 + 1; i++) {
			ack->bitmap[i] = 0xff;
		}
	} else {
		vector <u_char *> *pList = ci->getPacketList();
		for (i = 0; i < pktCount; i++) {
			hop = (struct hop_data *) MF_PacketSupport::getHopHeaderPtr((*pList)[i]);
			if (hop->pld_size == 0) { //did not get the packet yet
				ack->bitmap[i / 8] &= (0xfe << (i % 8));
				j++;
			}
		}
	}
	u_int z;
	//TODO remove the harcoded value
	char tempBitmap[512];
	for (z = 0; (z < (pktCount - 1) / 8 + 1) && z<512; z++) {
		sprintf(tempBitmap + z, "%x",ack->bitmap[z]);
	}
	MF_Log::mf_log(MF_DEBUG, "MF_Segmentor:sendAck sent bitmap: %s", tempBitmap);
	
	MF_IPProto::fillHeader(buffer, ETH_HEADER_SIZE, CSYN_SIZE + ((pktCount - 1) / 8 + 1), di->getSIP().c_str(), di->getAPIP().c_str());
	MF_EtherProto::fillHeader(buffer, 0, di->getSMAC().c_str(), di->getAPMAC().c_str());
	
	if ((ret = pcap_inject(mPcapHandle, (void *) buffer, ETH_HEADER_SIZE + IP_HEADER_SIZE + CSYN_SIZE + ((pktCount - 1) / 8 + 1))) < 0) {
		MF_Log::mf_log(MF_ERROR, "MF_Segmentor:sendAck pcap_inject() error, errno=%d", errno);
		return false;
	}
	MF_Log::mf_log(MF_DEBUG, "MF_Segmentor:sendAck Sent CSYN-ACK hop_ID=%d, recvd/lost/total %d/%d/%d if = %s", ntohl(ack->hop_ID), pktCount-j, j, pktCount, di->getName().c_str());
	
	//OML
#ifdef __OML__
	di->increaseOutBytesByInt(ETH_HEADER_SIZE + IP_HEADER_SIZE + CSYN_SIZE + ((pktCount - 1) / 8 + 1));
	di->increaseOutPackets();
#endif
	
	if (j == 0) { //get all the packets
		return true;
	}
	return false;
}

void MF_Segmentor::sendAssoRequest(u_int entityGUID, u_int hostGUID, u_short weight, MF_DeviceInfo *di) {
	MF_Log::mf_log(MF_DEBUG, "MF_Segmentor:sendAssoRequest send association request packet");
	u_char buffer[256];
	int ret = 0;
	struct client_assoc_pkt *asso;
	asso = (struct client_assoc_pkt *) MF_PacketSupport::getHopHeaderPtr(buffer);
	asso->type = htonl(ASSOC_PKT);
	asso->client_GUID = htonl(entityGUID);
	asso->host_GUID = htonl(hostGUID);
	asso->weight = htons(weight);
	MF_IPProto::fillHeader(buffer, ETH_HEADER_SIZE, sizeof(struct client_assoc_pkt), di->getSIP().c_str(), di->getAPIP().c_str());
	MF_EtherProto::fillHeader(buffer, 0, di->getSMAC().c_str(), di->getAPMAC().c_str());
	if ((ret = pcap_inject(mPcapHandle, (void *) buffer, ETH_HEADER_SIZE + IP_HEADER_SIZE + sizeof(struct client_assoc_pkt))) < 0) {
		MF_Log::mf_log(MF_ERROR, "pcap_sendpacket error, errno=%d", errno);
		return;
	}
	MF_Log::mf_log(MF_DEBUG, "MF_Segmentor: Sent Assoc pkt entity GUID = %u, host GUID = %u, if = %s weight = %u", entityGUID, hostGUID, di->getName().c_str(), weight);
	
#ifdef __OML__
	di->increaseOutBytesByInt(ETH_HEADER_SIZE + IP_HEADER_SIZE + sizeof(struct client_assoc_pkt));
	di->increaseOutPackets();
#endif
}

void MF_Segmentor::sendDassoRequest(u_int entityGUID, u_int hostGUID, MF_DeviceInfo *di) {
	MF_Log::mf_log(MF_DEBUG, "MF_Segmentor:sendDassoRequest send deassociation request packet");
	u_char buffer[256];
	int ret = 0;
	struct client_dassoc_pkt *asso;
	asso = (struct client_dassoc_pkt *) MF_PacketSupport::getHopHeaderPtr(buffer);
	asso->type = htonl(DASSOC_PKT);
	asso->entity_GUID = htonl(entityGUID);
	asso->host_GUID = htonl(hostGUID);
	MF_IPProto::fillHeader(buffer, ETH_HEADER_SIZE, sizeof(struct client_dassoc_pkt), di->getSIP().c_str(), di->getAPIP().c_str());
	MF_EtherProto::fillHeader(buffer, 0, di->getSMAC().c_str(), di->getAPMAC().c_str());
	if ((ret = pcap_inject(mPcapHandle, (void *) buffer, ETH_HEADER_SIZE + IP_HEADER_SIZE + sizeof(struct client_dassoc_pkt))) < 0) {
		MF_Log::mf_log(MF_ERROR, "pcap_sendpacket error, errno=%d", errno);
		return;
	}
	MF_Log::mf_log(MF_DEBUG, "MF_Segmentor: Sent d-asso request for entity GUID = %d, host GUID = %d, if = %s", entityGUID, hostGUID, di->getName().c_str());

#ifdef __OML__
	di->increaseOutBytesByInt(ETH_HEADER_SIZE + IP_HEADER_SIZE + sizeof(struct client_dassoc_pkt));
	di->increaseOutPackets();
#endif
}

void MF_Segmentor::sendLPAck(u_int sourceLPACK, u_int destinationLPACK, u_int seq_no_cp, MF_DeviceInfo *di){
	MF_Log::mf_log(MF_DEBUG, "MF_Segmentor:sendLPAck send deassociation request packet");
	u_char buffer[256];
	int ret;
	struct link_probe_ACK *lpa;
	lpa = (struct link_probe_ACK *)MF_PacketSupport::getHopHeaderPtr(buffer);
	lpa->type = htonl(LP_ACK_PKT);
	lpa->destinationLPACK = htonl(destinationLPACK);
	lpa->sourceLPACK = htonl(sourceLPACK);
	lpa->seq_no_cp = htonl(seq_no_cp);
	MF_IPProto::fillHeader(buffer, ETH_HEADER_SIZE, sizeof(struct link_probe_ACK), di->getSIP().c_str(), di->getAPIP().c_str());
	MF_EtherProto::fillHeader(buffer, 0, di->getSMAC().c_str(), di->getAPMAC().c_str());
	if ((ret = pcap_inject(mPcapHandle, (void *) buffer, ETH_HEADER_SIZE + IP_HEADER_SIZE + sizeof(struct link_probe_ACK))) < 0) {
		MF_Log::mf_log(MF_ERROR, "MF_DeviceManager:sendLPAck pcap_inject() error, errno=%d", errno);
	}
	else{
		MF_Log::mf_log(MF_DEBUG, "MF_DeviceManager:sendLPAck link prob ack sent");
	}
	
#ifdef __OML__
	di->increaseOutBytesByInt(ETH_HEADER_SIZE + IP_HEADER_SIZE + sizeof(struct link_probe_ACK));
	di->increaseOutPackets();
#endif
}
